import os.path

import pandas as pd
import pytest
import yaml
from snakemake.utils import validate

from bacpage.src import example


@pytest.fixture( scope="session" )
def project_directory( tmp_path_factory ) -> str:
    directory = tmp_path_factory.mktemp( "testing_directory" )
    project_directory = os.path.join( directory, "project" )
    example.create_project_directory( directory=project_directory )
    return project_directory


def test_creation_of_project_directory( project_directory: str ):
    assert os.path.exists( project_directory ), f"{project_directory} does not exist."
    assert os.path.isdir( project_directory ), f"{project_directory} exists but is not a directory."


def test_creation_of_input_directory( project_directory: str ):
    input_directory = os.path.join( project_directory, "input" )
    assert os.path.exists( input_directory ), f"{input_directory} does not exist."
    assert os.path.isdir( input_directory ), f"{input_directory} exists but is not a directory."


def test_config_file_exists( project_directory: str ):
    config_file = os.path.join( project_directory, "config.yaml" )
    assert os.path.exists( config_file ), f"{config_file} does not exist."
    assert os.path.isfile( config_file ), f"{config_file} exists but is not a file."
    assert os.path.getsize( config_file ) > 100, f"{config_file} should contain text but is empty."


def test_config_file_is_valid_yaml( project_directory: str ):
    config_file = os.path.join( project_directory, "config.yaml" )
    with open( config_file, "r" ) as parameters:
        project_yaml = yaml.safe_load( parameters )
        validate( project_yaml, "bacpage/schemas/Illumina_config.schema.yaml" )


def test_sample_file_exists( project_directory: str ):
    samples_file = os.path.join( project_directory, "sample_data.csv" )
    assert os.path.exists( samples_file ), f"{samples_file} does not exist."
    assert os.path.isfile( samples_file ), f"{samples_file} exists but is not a file."
    assert os.path.getsize( samples_file ) > 100, f"{samples_file} should contain text but is empty."


def test_sample_file_is_valid_csv( project_directory: str ):
    sf = pd.read_csv( os.path.join( project_directory, "sample_data.csv" ) )
    for col in ["sample", "read1", "read2"]:
        assert col in sf.columns, f"{col} not in columns."
    assert sf.shape == (3, 3)


def test_exits_if_project_directory_is_not_empty():
    with pytest.raises( SystemExit ):
        example.create_project_directory( directory="test/trees" )



import argparse
import sys
from pathlib import Path

import pandas as pd
import snakemake
from snakemake.utils import validate

from bacpage.src import common_funcs


def add_command_arguments( parser: argparse.ArgumentParser ):
    parser.description = "Assembles consensus sequence from raw sequencing reads."

    parser.add_argument(
        "directory", type=str, nargs="?", default=".", help="Path to valid project directory [current directory]."
    )
    parser.add_argument(
        "--denovo", action="store_true", help="Perform de novo assembly rather than reference-based assembly."
    )
    parser.add_argument(
        "--configfile", type=str, default=".", help="Path to assembly configuration file ['config.yaml']."
    )
    parser.add_argument(
        "--samples", type=str, default=".",
        help="Path to file detailing raw sequencing reads for all samples ['sample_data.csv']."
    )
    parser.add_argument(
        "--no-qa", action="store_true", help="Whether to skip quality assessment of assemblies [False]"
    )
    parser.add_argument( "--threads", type=int, default=-1, help="Number of threads available for assembly [all]." )
    parser.add_argument( "--verbose", action="store_true", help="Print lots of stuff to screen." )

    parser.set_defaults( command=assemble_entrypoint )


def postamble( denovo: bool, directory: Path ):
    if denovo:
        print()
        print( "Successfully performed de novo assembly of your samples" )
        print( f"Annotated assemblies are available at {directory / 'results/assembly/'}." )
        print(
            f"Quality metrics of your input data and assemblies are available at {directory / 'reports/qc_report.html'}. Open this file in a web browser to view." )
        print()
        print( "To perform antimicrobial resistance profiling, run `bacpage profile`." )
    else:
        print()
        print( "Successfully performed reference-based assembly of your samples." )
        print( f"Consensus sequences are available at {directory / 'results/consensus'}." )
        print(
            f"Quality metrics of your input data and consensus sequences are available at {directory / 'reports/qc_report.html'}. Open this file in a web browser to view." )
        print()
        print( "Generate a phylogenetic tree incorporating these samples using `bacpage phylogeny`. " )
        print( "Or, determine the presense of antimicrobial resistance genes using `bacpage profile`." )


def assemble_entrypoint( args: argparse.Namespace ):
    run_assemble(
        project_directory=args.directory,
        configfile=args.configfile,
        sample_data=args.samples,
        denovo=args.denovo,
        qc=not args.no_qa,
        threads=args.threads,
        verbose=args.verbose,
    )


def run_assemble( project_directory: str, configfile: str, sample_data: str, denovo: bool, qc: bool, threads: int,
                  verbose: bool = False ):
    # Check project directory
    project_directory = Path( project_directory ).absolute()
    assert project_directory.exists() and project_directory.is_dir(), f"Specified project directory {project_directory} does not exist. Please specify a valid directory."

    # Check config file
    print( "Loading and validating configuration file...", end="" )
    try:
        config = common_funcs.load_configfile( configfile, project_directory )
    except Exception:
        print( "Error" )
        raise
    print( "Done" )

    # Perform QC if enabled
    config["QC"] = qc

    # Check sample data
    print( "Loading and validating samples data...", end="" )
    try:
        metadata, skipped_samples = load_sampledata(
            sample_data, project_directory, check_size=config["preprocessing"]["check_size"],
            minimum_size=config["preprocessing"]["minimum_size"]
        )
    except Exception:
        print( "Error" )
        raise
    print( f"Found {len( metadata )} samples. " )

    if len( skipped_samples ) > 0:
        print( f"Skipping samples [{', '.join( skipped_samples )}] because they have no reads." )

    # Add samples dictionary to config, so snakemake can utilize it.
    config["SAMPLES"] = metadata.set_index( "sample" )[["read1", "read2"]].to_dict( orient="index" )

    # Calculate number of threads if not specified
    useable_threads = common_funcs.calculate_threads( threads )

    # Run the appropriate snakemake command
    config["DENOVO"] = denovo
    if denovo:
        snakefile = common_funcs.PACKAGE_DIR / "rules/denovo.smk"
    else:
        snakefile = common_funcs.PACKAGE_DIR / "rules/assemble.smk"
    assert snakefile.exists(), f"Snakefile does not exist. Checking {snakefile}"
    if verbose:
        status = snakemake.snakemake(
            snakefile, force_incomplete=True, workdir=project_directory,
            restart_times=common_funcs.RESTART_TIMES,
            config=config, cores=useable_threads, lock=False,
            quiet=False, printshellcmds=True, printreason=False
        )
    else:
        status = snakemake.snakemake(
            snakefile, force_incomplete=True, workdir=project_directory,
            restart_times=common_funcs.RESTART_TIMES,
            config=config, cores=useable_threads, lock=False,
            quiet=True, printshellcmds=False, printreason=False
        )
    if not status:
        sys.stderr.write( "Snakemake pipeline did not complete successfully. Check for error messages and rerun.\n" )
        sys.exit( -2 )

    postamble( denovo, project_directory )


def load_sampledata( specified_loc: str, project_directory: Path, check_size: bool = False,
                     minimum_size: int = 100 ) -> (pd.DataFrame, list):
    """ Attempts for find sample data using user supplied information. If sample data file is directly specified, use it, else
    search for the file in the project directory.

    Parameters
    ----------
    specified_loc : str
        Location of sample data file. Must be in CSV format. Use "." to attempt to automatically find file in project_direcctory.
    project_directory : pathlib.Path
        Location of project directory.
    check_size : bool
        Indicates whether to validate input file sizes.
    minimum_size
        Minimum file size for an input file to be considered valid. Not considered if check_size is False.

    Returns
    -------
    pd.DataFrame
        Dataframe matching sample names to input raw sequencing data.
    list
        Sample names which were removed from sample data because their input files are smaller than minimum_size.
    """
    sampledata_loc = Path( specified_loc )
    if specified_loc == ".":
        sampledata_loc = project_directory / common_funcs.DEFAULT_SAMPLEDATA
        if not sampledata_loc.exists():
            sys.stderr.write(
                f"Unable to automatically find sample data file in project directory (searching for '{common_funcs.DEFAULT_SAMPLEDATA}'). Please specify a valid sample data file."
            )
            sys.exit( -3 )
    elif not sampledata_loc.is_absolute():
        sampledata_loc = project_directory / sampledata_loc
    if not sampledata_loc.exists():
        sys.stderr.write( f"{sampledata_loc} does not exist. Please specify a valid file." )
        sys.exit( -4 )

    md = pd.read_csv( sampledata_loc )

    schema_location = common_funcs.PACKAGE_DIR / "schemas/Illumina_metadata.schema.yaml"
    validate( md, schema_location )

    md["read1"] = md["read1"].apply( lambda x: str( common_funcs.normalize_path( x, project_directory ) ) )
    md["read2"] = md["read2"].apply( lambda x: str( common_funcs.normalize_path( x, project_directory ) ) )

    duplicate_samples = md.loc[md["sample"].duplicated(), "sample"]
    if len( duplicate_samples ) > 0:
        sys.stderr.write(
            f"Sample data contains duplicate samples ({', '.join( duplicate_samples.to_list() )}) at rows {duplicate_samples.index.to_list()}"
        )
        sys.exit( -1 )

    skipped_samples = list()
    if check_size:
        md["size"] = md["read1"].apply( lambda x: Path( x ).stat().st_size )
        skipped_samples = md.loc[md["size"] < minimum_size, "sample"].to_list()
        md = md.loc[md["size"] >= minimum_size]

    return md, skipped_samples



import os
from pathlib import Path

import pandas as pd
import pytest

from bacpage.src import identify

SAMPLES = [
    "ERR976393",
    "MOZ-PMB0645689",
    "DRC-HL423",
    "CMR-CEN020JA",
    "ERR976399",
    "ERR976398"
]


@pytest.fixture( scope="session" )
def sample_data( tmp_path_factory ) -> str:
    directory = tmp_path_factory.mktemp( "testing_directory" )
    sample_data = os.path.join( directory, "sample_data.csv" )
    identify.generate_sample_data( "test/test_fastqs", sample_data )
    return sample_data


def test_if_exists( sample_data ):
    assert os.path.exists( sample_data ), f"{sample_data} does not exist."
    assert os.path.isfile( sample_data ), f"{sample_data} exists but is not a file."
    assert os.path.getsize( sample_data ) > 100, f"{sample_data} should contain text but is empty."


def test_if_valid_csv( sample_data ):
    sample_df = pd.read_csv( sample_data )
    for col in ["sample", "read1", "read2"]:
        assert col in sample_df.columns, f"{col} not in columns."
    assert sample_df.shape[1] > 0, "sample_data.csv contains no entries."


def test_if_all_samples_parsed( sample_data ):
    sample_df = pd.read_csv( sample_data )
    results = dict()
    for sample in SAMPLES:
        results[sample] = any( sample_df["sample"] == sample )
    not_found = [sample for sample in results if not results[sample]]
    assert len( not_found ) == 0, f"Not all samples found in samples_data.csv ({', '.join( not_found )})"


def test_if_sample_path_valid( sample_data ):
    sample_df = pd.read_csv( sample_data )
    for idx, entry in sample_df.iterrows():
        read1 = Path( entry["read1"] )
        read2 = Path( entry["read2"] )
        assert read1.exists(), f"{read1} does not exist."
        assert read1.is_absolute(), f"{read1} should be absolute but isn't."
        assert read2.exists(), f"{read2} does not exist."
        assert read2.is_absolute(), f"{read2} should be absolute but isn't."



import argparse
import sys
from pathlib import Path

import snakemake
from Bio import SeqIO

from bacpage.src import common_funcs

OTHER_IUPAC = {'r', 'y', 's', 'w', 'k', 'm', 'd', 'h', 'b', 'v'}
VALID_CHARACTERS = [{'a'}, {'c'}, {'g'}, {'t'}, {'n'}, OTHER_IUPAC, {'-'}, {'?'}]


def add_command_arguments( parser: argparse.ArgumentParser ):
    parser.description = "Reconstructs maximum likelihood phylogeny from consensus sequences."

    parser.add_argument(
        "directory", type=str, default=".", help="Path to valid project directory [current directory]."
    )
    parser.add_argument(
        "--configfile", type=str, default=".", help="Path to assembly configuration file ['config.yaml']."
    )
    parser.add_argument(
        "--minimum-completeness", type=float, default=0.9,
        help="minimum coverage required to be included in analysis [0.9]"
    )
    parser.add_argument( "--threads", type=int, default=-1, help="Number of threads available for command [all]." )
    parser.add_argument( "--verbose", action="store_true", help="Print lots of stuff to screen." )

    parser.set_defaults( command=phylogeny_entrypoint )


def calculate_completeness( sequence_loc: Path ) -> float:
    record = SeqIO.read( sequence_loc, "fasta" )
    seq = record.seq.lower()
    l = len( seq )
    counts = []

    for v in VALID_CHARACTERS:
        counts.append( sum( map( lambda x: seq.count( x ), v ) ) )
    invalid_nucleotides = l - sum( counts )

    if invalid_nucleotides > 0:
        print( "Invalid characters in sequence. Might not be a valid nucleotide sequence." )

    return 1.0 - (counts[4] / l)


def load_input( directory: str, minimum_completeness: float ) -> dict[str, Path]:
    search_directory = Path( directory ).absolute()
    if common_funcs.is_project_directory( search_directory ):
        search_directory = search_directory / "results/consensus"
    fastas = common_funcs.find_files( directory=search_directory, extensions=[".fa", ".fasta"] )

    if minimum_completeness > 0:
        for name in list( fastas.keys() ):
            completeness = calculate_completeness( fastas[name] )
            if completeness < minimum_completeness:
                del fastas[name]
    return fastas


def validate_sequences( sequence_paths: dict[str, Path], background_dataset: Path = None ):
    seen_names = list()

    if background_dataset and (background_dataset != ""):
        for record in SeqIO.parse( background_dataset, "fasta" ):
            seen_names.append( record.name )

    for path in sequence_paths.values():
        for record in SeqIO.parse( path, "fasta" ):
            seen_names.append( record.name )

    if len( seen_names ) != len( set( seen_names ) ):
        sys.stderr.write( "Sequence names are not unique. Remove duplicate sequences." )
        sys.exit( -6 )


def postamble( directory: Path ):
    print()
    print( f"Reconstructed phylogeny is available at {directory / 'results/phylogeny/phylogeny.tree'}" )
    print(
        "Open the file in any tree viewer to visualize. We recommend figtree (http://tree.bio.ed.ac.uk/software/figtree/)." )


def reconstruct_phylogeny( project_directory: str, configfile: str, minimum_completeness: float, threads: int,
                           verbose: bool ):
    project_path = Path( project_directory ).absolute()
    assert project_path.exists() and project_path.is_dir(), f"Specified project directory {project_path} does not exist. Please specify a valid directory."

    input_sequences = load_input( project_directory, minimum_completeness=minimum_completeness )

    # Check config file
    print( "Loading and validating configuration file...", end="" )
    try:
        config = common_funcs.load_configfile( configfile, project_path )
    except Exception:
        print( "Error" )
        raise
    print( "Done" )

    if config["background_dataset"] not in ["", "<background-dataset-path>"]:
        background_dataset = Path( config["background_dataset"] ).absolute()
        config["BACKGROUND"] = background_dataset
    else:
        config["BACKGROUND"] = ""

    validate_sequences( input_sequences, config["BACKGROUND"] )

    config["SAMPLES"] = input_sequences

    # Calculate number of threads
    useable_threads = common_funcs.calculate_threads( threads )

    # load snakefile
    snakefile = common_funcs.PACKAGE_DIR / "rules/phylogeny.smk"
    assert snakefile.exists(), f"Snakefile does not exist. Searching for {snakefile}"
    status = snakemake.snakemake(
        snakefile, printshellcmds=True, forceall=True, force_incomplete=True, workdir=project_directory,
        restart_times=common_funcs.RESTART_TIMES, config=config, cores=useable_threads, lock=False, quiet=not verbose
    )
    if not status:
        sys.stderr.write( "Snakemake pipeline did not complete successfully. Check for error messages and rerun." )
        sys.exit( -2 )

    postamble( project_directory )


def phylogeny_entrypoint( args: argparse.Namespace ):
    reconstruct_phylogeny(
        project_directory=args.directory,
        configfile=args.configfile,
        minimum_completeness=args.mininum_completeness,
        threads=args.threads,
        verbose=args.verbose,
    )



import sys
from importlib.resources import files
from pathlib import Path

import yaml
from snakemake import WorkflowError
from snakemake.utils import validate

DEFAULT_CONFIG = "config.yaml"
DEFAULT_SAMPLEDATA = "sample_data.csv"
PACKAGE_DIR = files( "bacpage" )
CONFIG_PATHS = ["reference", "reference_genes", "recombinant_mask"]
RESTART_TIMES = 0


def find_files( directory: Path, extensions: list[str] ) -> dict[str, Path]:
    search_path = directory.absolute()
    found = dict()
    for file in search_path.iterdir():
        if file.suffix in extensions:
            name = file.stem
            if name in found:
                sys.stderr.write( f"{name} was found in two or more files. Duplicate sequences cannot be processed." )
                sys.stderr.write( "fuck off" )
                sys.exit( -5 )
            path = search_path / file
            found[name] = path
    return found


def is_project_directory( directory ):
    return (directory / "sample_data.csv").exists() and (directory / "config.yaml").exists()


def normalize_path( value: str, working_directory: Path ) -> Path:
    path = Path( value )
    if not path.is_absolute():
        return working_directory / path
    else:
        return path


def load_configfile( specified_loc: str, project_directory: Path ) -> dict:
    """ Attempts for find config file using user supplied information. If config file is directly specified, use it, else
    search for the config file in the project directory.

    Parameters
    ----------
    specified_loc: str
        Path to config file. Pass "." to automatically search for file in project directory.
    project_directory: pathlib.Path
        Path to project directory. Used if config file path is not specified and to normalize relative paths in the config file.

    Returns
    -------
    dict
        Config file loaded as a python object.
    """
    configfile_loc = Path( specified_loc ).absolute()
    if specified_loc == ".":
        configfile_loc = project_directory / DEFAULT_CONFIG
        assert configfile_loc.exists(), "Unable to automatically find config in project directory (searching for 'config.yaml'). Please specify a valid configuration file."
    assert configfile_loc.exists(), f"{configfile_loc} does not exist. Please specify a valid file."

    with open( configfile_loc, "r" ) as cf:
        configfile = yaml.safe_load( cf )

    schema_location = PACKAGE_DIR / "schemas/Illumina_config.schema.yaml"

    try:
        validate( configfile, schema_location )
    except WorkflowError as err:
        message = err.args[0].split( "\n" )
        print( "Error" )
        sys.stderr.write( f"{message[0]} {message[1].split( ': ' )[1]}.\n" )
        sys.exit( -9 )

    for key in CONFIG_PATHS:
        configfile[key] = str( normalize_path( configfile[key], PACKAGE_DIR / "resources" ) )

    return configfile


def calculate_threads( threads ):
    if threads == 0:
        sys.stderr.write(
            "Pipeline cannot function without threads. Please specify a non-zero number of threads with the '--threads' "
            "option or run comman without '--threads' option to automatically detected the number of available threads."
        )
        sys.exit( -4 )
    elif threads < 0:
        import multiprocessing
        threads = multiprocessing.cpu_count()
    print( f"Using {threads} threads." )

    return threads



import shutil
from pathlib import Path

import pytest
from Bio import Phylo

from bacpage.src import phylogeny


def test_recognize_folder_of_fastas():
    search_directory = "test/test_tree_fasta_directory"
    found = phylogeny.load_input( directory=search_directory, minimum_completeness=0 )
    search_directory = Path( search_directory ).absolute()
    expected = [search_directory / f"t{i}.fasta" for i in range( 1, 21 )]
    assert sorted( found.values() ) == sorted( expected )


def test_recognize_project_directory():
    search_directory = "test/test_tree_project_directory"
    found = phylogeny.load_input( directory=search_directory, minimum_completeness=0 )
    search_directory = Path( search_directory ).absolute() / "results/consensus"
    expected = [search_directory / f"t{i}.fasta" for i in range( 1, 21 )]
    assert sorted( found.values() ) == sorted( expected )


def test_remove_low_coverage_sequences():
    search_directory = "test/test_tree_fasta_directory"
    found = phylogeny.load_input( directory=search_directory, minimum_completeness=0.9 )
    search_directory = Path( search_directory ).absolute()
    expected = [search_directory / f"t{i}.fasta" for i in range( 1, 15 )]
    assert sorted( found.values() ) == sorted( expected )


def test_raise_error_with_duplicate_samples_in_directory():
    search_directory = "test/test_tree_duplicate_samples"
    with pytest.raises( SystemExit ) as excinfo:
        found = phylogeny.load_input( directory=search_directory, minimum_completeness=0.9 )
    assert excinfo.value.code == -5


def test_raise_error_with_duplicate_sample_names():
    search_directory = "test/test_tree_duplicate_sequences"
    with pytest.raises( SystemExit ) as excinfo:
        phylogeny.reconstruct_phylogeny( search_directory, "test/test_tree_fasta_directory/config.yaml", 0, 1, True )
    assert excinfo.value.code == -6


@pytest.fixture()
def phylogeny_run( scope="session" ):
    project_directory = Path( "test/test_tree_fasta_directory" )
    phylogeny.reconstruct_phylogeny( str( project_directory ), ".", minimum_completeness=0.9, threads=-1,
                                     verbose=False )
    yield project_directory

    if (project_directory / "results").exists():
        shutil.rmtree( project_directory / "results" )
    if (project_directory / "intermediates").exists():
        shutil.rmtree( project_directory / "intermediates" )


def test_phylogeny_postamble():
    project_directory = Path( "test/test_tree_fasta_directory" ).absolute()
    phylogeny.postamble( project_directory )


@pytest.mark.slow
def test_tree_reconstruction_successfully( phylogeny_run ):
    tree = phylogeny_run / "results/phylogeny/phylogeny.tree"
    assert tree.exists() and tree.is_file(), "Phylogeny was either not created or is not a file."


@pytest.mark.slow
def test_tree_reconstruction_all_taxa_present( phylogeny_run ):
    tree = phylogeny_run / "results/phylogeny/phylogeny.tree"
    tree = Phylo.read( tree, "newick" )

    found = [clade.name for clade in tree.get_terminals()]
    expected = [f"t{i}" for i in range( 1, 15 )]

    assert sorted( found ) == sorted( expected ), "Not all taxa where found in tree."
    assert 1 < tree.total_branch_length() < 100, f"Branch length ({tree.total_branch_length()}) is not a reasonable magnitude."



import shutil
from pathlib import Path

import pandas as pd
import pytest
from snakemake.utils import validate

from bacpage.src import assemble, common_funcs


def test_error_if_project_not_found():
    with pytest.raises( AssertionError ):
        assemble.run_assemble(
            project_directory="/foo/bar",
            configfile=".",
            sample_data=".",
            denovo=False,
            qc=True,
            threads=1
        )


def test_exit_if_zero_threads_allocated():
    with pytest.raises( SystemExit ) as excinfo:
        assemble.run_assemble(
            project_directory="test/test_pipeline",
            configfile=".",
            sample_data=".",
            denovo=False,
            qc=True,
            threads=0
        )


def test_error_if_config_not_automatically_found():
    with pytest.raises( AssertionError ):
        common_funcs.load_configfile( ".", Path( "test/test_fastqs" ).absolute() )


def test_automatically_find_config_file():
    config = common_funcs.load_configfile( ".", Path( "test/test_pipeline" ).absolute() )
    assert config, "An object should be returned but nothing was."


def test_config_is_valid_yaml():
    config = common_funcs.load_configfile( ".", Path( "test/test_pipeline" ).absolute() )
    assert isinstance( config, dict ), f"Returned config is a {type( config )}, expected <class 'dict'>."
    validate( config, "bacpage/schemas/Illumina_config.schema.yaml" )


def test_config_finds_local_parameters():
    config = common_funcs.load_configfile(
        "test/configs/local_parameters.yaml", Path( "test/test_pipeline" ).absolute()
    )
    reference = Path( config["reference"] )
    assert reference.is_absolute() and reference.exists(), f"Local reference in resources directory was not correctly found ({reference})."
    reference_genes = Path( config["reference_genes"] )
    assert reference_genes.is_absolute() and reference_genes.exists(), f"Local reference gene directory was not correctly found ({reference_genes})."
    recombinant_mask = Path( config["recombinant_mask"] )
    assert recombinant_mask.is_absolute() and recombinant_mask.exists(), f"Local reference gene directory was not correctly found ({recombinant_mask})."


def test_error_if_sample_data_not_automatically_found():
    with pytest.raises( SystemExit ) as excinfo:
        assemble.load_sampledata( ".", Path( "test/test_fastqs" ).absolute() )
    assert excinfo.value.code == -3


def test_automatically_find_sample_data():
    sample_data, skipped_samples = assemble.load_sampledata( ".", Path( "test/test_pipeline" ).absolute() )
    assert sample_data is not None, "An object should be returned but nothing was."


def test_sample_data_is_valid_dataframe():
    sample_data, skipped_samples = assemble.load_sampledata( ".", Path( "test/test_pipeline" ).absolute() )
    assert isinstance(
        sample_data, pd.DataFrame
    ), f"Returned sample data is {type( sample_data )}, expected <class 'pd.DataFrame'>."
    validate( sample_data, "bacpage/schemas/Illumina_metadata.schema.yaml" )


def test_duplicate_samples_causes_exit():
    with pytest.raises( SystemExit ) as excinfo:
        sample_data, skipped_samples = assemble.load_sampledata(
            "duplicate_names.csv", Path( "test/sample_datas" ).absolute()
        )
    assert excinfo.value.code == -1


@pytest.mark.slow
def test_assemble_snakemake_runs_correctly():
    project_directory = Path( "test/test_pipeline/" )
    expected_output = ["results/consensus/test.consensus.fasta", "results/reports/depth/test.depth.pdf",
                       "results/reports/qc_report.html"]
    expected_output = [project_directory / file for file in expected_output]

    assemble.run_assemble(
        project_directory=str( project_directory ),
        configfile=".",
        sample_data=".",
        denovo=False,
        qc=True,
        threads=-1,
        verbose=False
    )
    results = dict()
    for file in expected_output:
        results[file] = file.exists()
    assert all( results.values() ), f"Not all expected files generated. Expected {results}"

    if (project_directory / "results").exists():
        shutil.rmtree( project_directory / "results" )
    if (project_directory / "intermediates").exists():
        shutil.rmtree( project_directory / "intermediates" )


def test_assemble_postamble():
    project_directory = Path( "test/test_pipeline/" ).absolute()
    assemble.postamble( False, project_directory )


@pytest.mark.slow
def test_denono_assembly_snakemake_runs_correctly():
    project_directory = Path( "test/test_pipeline/" )
    expected_output = ["results/assembly/test.annotated.gff", "results/reports/qc_report.html"]
    expected_output = [project_directory / file for file in expected_output]

    assemble.run_assemble(
        project_directory=str( project_directory ),
        configfile=".",
        sample_data=".",
        denovo=True,
        qc=True,
        threads=8
    )
    results = dict()
    for file in expected_output:
        results[file] = file.exists()
    assert all( results.values() ), f"Not all expected files generated. Expected {results}"

    if (project_directory / "results").exists():
        shutil.rmtree( project_directory / "results" )
    if (project_directory / "intermediates").exists():
        shutil.rmtree( project_directory / "intermediates" )

# def test_find_all_genes():
#    genes_found = list( assemble.get_genes( "resources/cholera_ref_genes/" ).keys() )
#    genes_expected = ["ctxA", "tcpA_classical", "tcpA_eltor", "toxR", "wbeO1", "wbfO139"]
#    assert sorted( genes_found ) == sorted(
#        genes_expected
#    ), f"Incorrect genes found. Expected {genes_expected}, got {genes_found}."
#
#
# def test_error_on_empty_directory():
#    with pytest.raises( AssertionError ):
#        genes_found = list( assemble.get_genes( "test" ).keys() )
#



import argparse
import pandas as pd
import matplotlib.pyplot as plt
import matplotlib.ticker as mticker
from matplotlib.lines import Line2D

if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Description"
    )

    # Initialize optional arguments
    parser.add_argument( "--input", type=str, help="Input depth file. " )
    parser.add_argument( "--output", type=str, help="output" )
    parser.add_argument( "--min-depth", type=int, help="Minimum required depth to call a position" )
    parser.add_argument( "--bin-size", type=int, help="Width of bins to " )

    args = parser.parse_args()

    depth = pd.read_csv( args.input, names=["chromosome", "position", "depth"], sep="\t" )
    depth["bin"] = (depth["position"] / args.bin_size).astype( int )
    depth = depth.groupby( ["chromosome", "bin"] )["depth"].agg( ["mean", "max", "min"] )
    depth = depth.reset_index()

    min_depth = args.min_depth

    chromosomes = depth["chromosome"].unique()

    fig, ax = plt.subplots( dpi=200, figsize=(10, 4 * len( chromosomes )), nrows=len( chromosomes ), sharey=True )
    ticks = mticker.FuncFormatter( lambda x, pos: '{0:,.0f}'.format( x * args.bin_size ) )

    leg = [
        Line2D( [0], [0], linestyle='-', marker=None, label=f"Average reads", markersize=0, linewidth=1 ),
        Line2D(
            [0], [0], linestyle='--', marker=None, color="black", label=f"Required depth ({min_depth})", markersize=0,
            linewidth=1
        )
    ]

    ax[0].legend( handles=leg, loc="upper left", handletextpad=1, frameon=False, fontsize=10, handlelength=1 )

    for i, name in enumerate( chromosomes ):
        data = depth.loc[depth["chromosome"] == name]
        ax[i].plot( "bin", "mean", data=data, zorder=10 )
        ax[i].fill_between( "bin", "min", "max", alpha=0.2, data=data, zorder=5 )
        ax[i].axhline( y=min_depth, color="black", linestyle="--", linewidth=1 )
        ax[i].xaxis.set_major_formatter( ticks )
        ax[i].set_xlabel( "Position (bp)" )
        ax[i].set_ylabel( "Reads" )
        ax[i].set_xlim( 0 )
        ax[i].set_ylim( 0 )
        ax[i].set_title( name )
        ax[i].grid( which="both", axis="both", linewidth=1, color="#F1F1F1", zorder=1 )

    plt.tight_layout()
    plt.savefig( args.output )
